#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <cmqc.h>
#include <cmqpsc.h>

#define MSGSIZ 1000
#define BFRSIZ 2000

#define S_REGISTER    1
#define P_REGISTER    2
#define S_DEREGISTER  3
#define P_DEREGISTER  4
#define PUBLISH       5
#define DELETE_PUB    6
#define REQUEST_UPD   7

#define REPLY_OUTPUT_FILE "outf"

#define REGOPT_CORRELID     0x0001
#define REGOPT_DEREGALL     0x002
#define REGOPT_NEWPUBS      0x0004
#define REGOPT_PUBONREQ     0x0008
#define REGOPT_INFORMIFRET  0x0010
#define PUBOPT_RETAINPUB    0x0001

#define ERR_NOQMGR  -1
#define ERR_NOQUEUE -2
#define ERR_RETPUB  -3
#define ERR_REGSUB  -4
#define ERR_DEREG   -5
#define ERR_TOPIC   -6

/**************************************************

Change history
--------------
1.0 
  Initial version
1.1 
  Register Subscriber and Deregister Subscriber TESTED OK
  Add support for CorrelId in text and binary
1.2
  Add reply-to queue, and get reply msg from broker
1.3
  Add more commands and options

**************************************************/

#define PROGVER "1.3"

int main(int argc, char **argv)
{
  /*  Declare file and character for sample input                   */
  FILE *fp;

  /*   Declare MQI structures needed                                */
  MQOD     od  = {MQOD_DEFAULT};    /* Object Descriptor             */
  MQOD     odr = {MQOD_DEFAULT};    /* Object Descriptor Reply       */
  MQMD     md  = {MQMD_DEFAULT};    /* Message Descriptor            */
  MQPMO    pmo = {MQPMO_DEFAULT};   /* put message options           */
  MQGMO    gmo = {MQGMO_DEFAULT};   /* get message options           */
  MQRFH    mqrfh = {MQRFH_DEFAULT};
  PMQRFH   pmqrfh = &mqrfh;
  char    *pnv;

  MQHCONN  Hcon;                   /* connection handle             */
  MQHOBJ   Hobj;                   /* object handle                 */
  MQHOBJ   Hobjr;                  /* object handle reply           */
  MQLONG   O_options;              /* MQOPEN options                */
  MQLONG   C_options;              /* MQCLOSE options               */
  MQLONG   CompCode;               /* completion code               */
  MQLONG   OpenCode;               /* MQOPEN completion code        */
  MQLONG   Reason;                 /* reason code                   */
  MQLONG   CReason;                /* reason code for MQCONN        */
  MQLONG   messlen;                /* message length                */
  MQLONG   buflen;                 /* buffer length                 */
  char     msgdata[MSGSIZ];        /* message buffer                */
  char     buffer[BFRSIZ];         /* message buffer                */
  char     QMName[50];             /* queue manager name            */
  char     OQName[50];             /* output queue name             */
  char     DQName[50];             /* destination queue name */
  char     RQName[50];             /* reply queue name */
  char     topic[100];
  char     stream[100];
  char     CorrelId[30];
  int      doRegOpts = 0;
  int      doPubOpts = 0;
  int      ok=1, opt, action = 0;
  char     fname[100];             /* filename of message to publish */
  char     rname[100];             /* reply output file name         */
  int      getReply=0;

  printf("Example %s %s start\n",argv[0],PROGVER);

  strcpy(OQName,"SYSTEM.BROKER.CONTROL.QUEUE");
  strcpy(QMName,"");
  strcpy(fname,"");
  strcpy(rname,REPLY_OUTPUT_FILE);
  strcpy(topic,"");
  strcpy(stream,"SYSTEM.BROKER.DEFAULT.STREAM");
  strcpy(DQName,"");
  strcpy(RQName,"");
  strcpy(CorrelId,"");

  while (ok && (opt = getopt(argc, argv, "rRdDpPUanoikq:t:s:f:M:Q:F:c:C:")) != -1)
  {
    switch (opt)
    {
    /* CONNECTION */
    case 'M': /* qmgr */
      strcpy(QMName,optarg);
      break;
    case 'Q': /* reply-to queue name */
      strcpy(RQName,optarg);
      break;
    case 'F': /* reply output file name */
      strcpy(rname,optarg);
      break;
    /* ACTIONS */
    case 'r': /* register subscriber */
      if (action != 0)
        ok = 0;
      action |= S_REGISTER;
      break;
    case 'R': /* register publisher */
      if (action != 0)
        ok = 0;
      action |= P_REGISTER;
      break;
    case 'd': /* deregister subscriber */
      if (action != 0)
        ok = 0;
      action |= S_DEREGISTER;
      break;
    case 'D': /* deregister publisher */
      if (action != 0)
        ok = 0;
      action |= P_DEREGISTER;
      break;
    case 'p': /* publish */
      if (action != 0)
        ok = 0;
      action |= PUBLISH;
      break;
    case 'P': /* delete publication */
      if (action != 0)
        ok = 0;
      action |= DELETE_PUB;
      break;
    case 'U': /* request update */
      if (action != 0)
        ok = 0;
      action |= REQUEST_UPD;
      break;
    /* DATA */
    case 't': /* topic */
      strcpy(topic,optarg);
      break;
    case 's': /* stream */
      strcpy(stream,optarg);
      break;
    case 'q': /* destination queue name */
      strcpy(DQName,optarg);
      break;
    case 'f': /* input file name */
      strcpy(fname,optarg);
      break;
    /* OPTIONS */
    case 'c': /* CorrelId text */
      if (strlen(optarg) > 24)
      {
	fprintf(stderr,"CorrelId text too long, max 24 chars");
	fflush(stderr);
	ok = 0;
      }
      else
      {
	memset(CorrelId,' ',24);
	strncpy(CorrelId,optarg,strlen(optarg));
	doRegOpts |= REGOPT_CORRELID;
      }
      break;
    case 'C': /* CorrelId binary chars */
      if (strlen(optarg) > 48)
      {
	fprintf(stderr,"CorrelId binary data too long, max 48 chars");
	fflush(stderr);
	ok = 0;
      }
      else
      {
	int l;
	l = txthex(optarg,strlen(optarg),CorrelId);
	while (l < 24)
	{
	  CorrelId[l++] = '\0';
	}
	if (0)
	{
	  for (l = 0; l < 24; l++)
	  {
	    printf("%02X",CorrelId[l]);
	    if (l && l%4 == 0)
	    printf(" ");
	  }
	}
	doRegOpts |= REGOPT_CORRELID;
      }
      break;
    case 'a': /* Deregister All */
      doRegOpts |= REGOPT_DEREGALL;
      break;
    case 'n': /* New Pubs Only */
      doRegOpts |= REGOPT_NEWPUBS;
      break;
    case 'o': /* Pub On Req Only */
      doRegOpts |= REGOPT_PUBONREQ;
      break;
    case 'i': /* Inform if Ret */
      doRegOpts |= REGOPT_INFORMIFRET;
      break;
    case 'k': /* Retain Publication */
      doPubOpts |= PUBOPT_RETAINPUB;
      break;
    case '?':
      ok = 0;
      break;
    }
  }

  if (action == 0)
    ok = 0;

  if (ok && *QMName == '\0')
    ok = ERR_NOQMGR;

  if (ok && *DQName == '\0' && action != DELETE_PUB)
    ok = ERR_NOQUEUE;

  if (ok && (doPubOpts & PUBOPT_RETAINPUB) && action != PUBLISH)
    ok = ERR_RETPUB;

  if (ok && (doRegOpts & (REGOPT_NEWPUBS | REGOPT_PUBONREQ | REGOPT_INFORMIFRET) != 0) 
         && action != S_REGISTER
     )
    ok = ERR_REGSUB;

  if (ok && (doRegOpts & REGOPT_DEREGALL) 
         && action != S_DEREGISTER 
         && action != P_DEREGISTER 
         && *topic != '\0'
     )
    ok = ERR_DEREG;

  if (ok && *topic == '\0' && (doRegOpts & REGOPT_DEREGALL) == 0)
    ok = ERR_TOPIC;

  if (ok <= 0)
  {
    usage(argv[0],ok);
    exit(1);
  }

  MQCONN(QMName,
        &Hcon, 
        &CompCode,
        &CReason);

  /* report reason and stop if it failed     */
  if (CompCode == MQCC_FAILED)
  {
   printf("MQCONN ended with reason code %ld\n", CReason);
   exit( (int)CReason );
  }

  /* publish puts to stream queue not control queue */
  if (action == PUBLISH)
    strcpy(OQName,stream);
  strncpy(od.ObjectName, OQName, (size_t)MQ_Q_NAME_LENGTH);
  printf("target queue is %s\n", od.ObjectName);

  O_options = MQOO_OUTPUT           /* open queue for output        */
         + MQOO_FAIL_IF_QUIESCING; /* but not if MQM stopping      */
  MQOPEN(Hcon,                      /* connection handle            */
        &od,                       /* object descriptor for queue  */
        O_options,                 /* open options                 */
        &Hobj,                     /* object handle                */
        &OpenCode,                 /* MQOPEN completion code       */
        &Reason);                  /* reason code                  */

  /* report reason, if any; stop if failed      */
  if (Reason != MQRC_NONE)
  {
    printf("MQOPEN ended with reason code %ld\n", Reason);
  }

  if (OpenCode == MQCC_FAILED)
  {
    printf("unable to open queue for output\n");
  }
  else
  {
    /* Open ReplyToQ if required */
    if (*RQName != '\0')
    {
      getReply = 1;
      strncpy(odr.ObjectName, RQName, (size_t)MQ_Q_NAME_LENGTH);

      O_options = MQOO_INPUT_AS_Q_DEF + MQOO_FAIL_IF_QUIESCING;
      MQOPEN(Hcon,                      /* connection handle            */
	    &odr,                       /* object descriptor for queue  */
	    O_options,                 /* open options                 */
	    &Hobjr,                     /* object handle                */
	    &OpenCode,                 /* MQOPEN completion code       */
	    &Reason);                  /* reason code                  */

      /* report reason, if any; stop if failed      */
      if (Reason != MQRC_NONE)
      {
	printf("MQOPEN ended with reason code %ld\n", Reason);
      }
    }
  }

  CompCode = OpenCode;        /* use MQOPEN result for initial test */
  memcpy(md.Format, MQFMT_RF_HEADER, (size_t)MQ_FORMAT_LENGTH);

  if (*fname == '\0')
    fp = stdin;
  else
  {
    fp = fopen(fname,"r");
    if (fp == NULL)
      CompCode = MQCC_FAILED;
  }

  if (action == PUBLISH)
    messlen = get_message(fp,msgdata);
  else
    messlen = 0;

  if (CompCode == MQCC_OK && messlen >= 0)
  {
    long pnvl;

    memset(buffer,0,BFRSIZ);
    memcpy(buffer,pmqrfh,sizeof(MQRFH));
    pmqrfh = (PMQRFH)buffer;
    pnv = (char *)(pmqrfh+1);
    pmqrfh->CodedCharSetId = MQCCSI_INHERIT;

    /* command string */
    strcpy(pnv,MQPS_COMMAND);

    /* command name */
    switch (action)
    {
    case P_REGISTER:
      strcat(pnv,MQPS_REGISTER_PUBLISHER_B);
      break;
    case S_REGISTER:
      strcat(pnv,MQPS_REGISTER_SUBSCRIBER_B);
      break;
    case S_DEREGISTER:
      strcat(pnv,MQPS_DEREGISTER_SUBSCRIBER_B);
      break;
    case P_DEREGISTER:
      strcat(pnv,MQPS_DEREGISTER_PUBLISHER_B);
      break;
    case PUBLISH:
      strcat(pnv,MQPS_PUBLISH_B);
      break;
    case DELETE_PUB:
      strcat(pnv,MQPS_DELETE_PUBLICATION_B);
      break;
    case REQUEST_UPD:
      strcat(pnv,MQPS_DELETE_PUBLICATION_B);
      break;
    }

    /*
      stream
      qmgr name
      queue name
    */

    strcat(pnv,MQPS_STREAM_NAME_B);
    strcat(pnv,stream);

    strcat(pnv,MQPS_Q_MGR_NAME_B);
    strcat(pnv,QMName);

    if (*DQName !='\0')
    {
      strcat(pnv,MQPS_Q_NAME_B);
      strcat(pnv,DQName);
    }

    /*
      Registration options implemented so far:
        CorrelAsId
    */

    if (doRegOpts)
    {
      if (doRegOpts & REGOPT_CORRELID)
      {
	strcat(pnv,MQPS_REGISTRATION_OPTIONS_B);
	strcat(pnv,MQPS_CORREL_ID_AS_IDENTITY_B);
      }
      if (doRegOpts & REGOPT_DEREGALL)
      {
	strcat(pnv,MQPS_REGISTRATION_OPTIONS_B);
	strcat(pnv,MQPS_DEREGISTER_ALL_B);
      }
      if (doRegOpts & REGOPT_NEWPUBS)
      {
	strcat(pnv,MQPS_REGISTRATION_OPTIONS_B);
	strcat(pnv,MQPS_NEW_PUBLICATIONS_ONLY_B);
      }
      if (doRegOpts & REGOPT_PUBONREQ)
      {
	strcat(pnv,MQPS_REGISTRATION_OPTIONS_B);
	strcat(pnv,MQPS_PUBLISH_ON_REQUEST_ONLY_B);
      }
      if (doRegOpts & REGOPT_INFORMIFRET)
      {
	strcat(pnv,MQPS_REGISTRATION_OPTIONS_B);
	strcat(pnv,MQPS_INFORM_IF_RETAINED_B);
      }
    }

    if (doPubOpts)
    {
      if (doPubOpts & PUBOPT_RETAINPUB)
      {
	strcat(pnv,MQPS_PUBLICATION_OPTIONS_B);
        strcat(pnv,MQPS_RETAIN_PUBLICATION_B);
      }
    }

    /*
      topic
      enclosed in double quotes if the topic contains a space
    */

    if (*topic != '\0')
    {
      strcat(pnv,MQPS_TOPIC_B);
      if (strchr(topic,' ') != NULL)
      {
	strcat(pnv,"\"");
	strcat(pnv,topic);
	strcat(pnv,"\"");
      }
      else
	strcat(pnv,topic);
    }

    printf("Name/Value pairs <%s>\n",pnv);

    pnvl = strlen(pnv);
    if (pnvl %4 != 0)
      pnvl += 4 - pnvl % 4;
    pmqrfh->StrucLength += pnvl;

    strncpy(buffer + pmqrfh->StrucLength,msgdata,messlen);
    messlen += pmqrfh->StrucLength;

    memcpy(md.MsgId,           /* reset MsgId to get a new one    */
          MQMI_NONE, sizeof(md.MsgId) );

    if (doRegOpts & REGOPT_CORRELID)
    {
      memcpy(md.CorrelId,CorrelId,24);
    }
    else
    {
      memcpy(md.CorrelId,        /* reset CorrelId to get a new one */
          MQCI_NONE, sizeof(md.CorrelId) );
    }

    if (getReply)
    {
      strncpy(md.ReplyToQ,RQName,strlen(RQName));
      md.MsgType = MQMT_REQUEST;
    }

    MQPUT(Hcon,                /* connection handle               */
         Hobj,                /* object handle                   */
         &md,                 /* message descriptor              */
         &pmo,                /* default options (datagram)      */
         messlen,             /* message length                  */
         buffer,              /* message buffer                  */
         &CompCode,           /* completion code                 */
         &Reason);            /* reason code                     */

    /* report reason, if any */
    if (Reason != MQRC_NONE)
    {
      printf("MQPUT ended with reason code %ld\n", Reason);
    }
    else
    {
      if (getReply)
      {
        FILE *fpo;
        MQMD  gmd = {MQMD_DEFAULT};
  
        memcpy(gmd.MsgId,md.MsgId,24);

        fpo = fopen(rname,"a");
        gmo.Version = MQGMO_VERSION_2;
        gmo.MatchOptions = MQMO_MATCH_CORREL_ID;
        gmo.Options = MQGMO_WAIT + MQGMO_CONVERT;
        gmo.WaitInterval = 2000;

        buflen = sizeof(buffer) - 1; /* buffer size available for GET   */

        gmd.Encoding       = MQENC_NATIVE;
        gmd.CodedCharSetId = MQCCSI_Q_MGR;
  
        MQGET(Hcon,                /* connection handle                 */
              Hobjr,                /* object handle                     */
              &gmd,                 /* message descriptor                */
              &gmo,                /* get message options               */
              buflen,              /* buffer length                     */
              buffer,              /* message buffer                    */
              &messlen,            /* message length                    */
              &CompCode,           /* completion code                   */
              &Reason);            /* reason code                       */

        /* report reason, if any     */
        if (Reason != MQRC_NONE)
        {
          if (Reason == MQRC_NO_MSG_AVAILABLE)
          {                         /* special report for normal end    */
            printf("no more messages\n");
          }
          else                      /* general report for other reasons */
          {
            printf("MQGET ended with reason code %ld\n", Reason);

            /*   treat truncated message as a failure for this sample   */
            if (Reason == MQRC_TRUNCATED_MSG_FAILED)
            {
              CompCode = MQCC_FAILED;
            }
          }
        }
        else
        {
          printf("reply message retrieved\n");
          fwrite(buffer,1,messlen,fpo);
          dbgprt(stdout,buffer,messlen);
        }
          fclose(fpo);
      }
    }
  }

  if (OpenCode != MQCC_FAILED)
  {
    C_options = 0;                  /* no close options             */
    MQCLOSE(Hcon,                   /* connection handle            */
           &Hobj,                  /* object handle                */
           C_options,
           &CompCode,              /* completion code              */
           &Reason);               /* reason code                  */

    /* report reason, if any     */
    if (Reason != MQRC_NONE)
    {
      printf("MQCLOSE ended with reason code %ld\n", Reason);
    }
    if (getReply)
    {
      MQCLOSE(Hcon,                   /* connection handle            */
             &Hobjr,                  /* object handle                */
             C_options,
             &CompCode,              /* completion code              */
             &Reason);               /* reason code                  */

      /* report reason, if any     */
      if (Reason != MQRC_NONE)
      {
        printf("MQCLOSE ended with reason code %ld\n", Reason);
      }
    }
  }

  if (CReason != MQRC_ALREADY_CONNECTED)
  {
    MQDISC(&Hcon,                   /* connection handle            */
            &CompCode,               /* completion code              */
            &Reason);                /* reason code                  */

    /* report reason, if any     */
    if (Reason != MQRC_NONE)
    {
      printf("MQDISC ended with reason code %ld\n", Reason);
    }
  }

  printf("Sample %s end\n",argv[0]);
  return(0);
}

txthex(char *indata, int len, char *outdata)
{
  int j=0, k=0, ch=0,fst=1;
  int c;

  while (j < len)
  {
    c = indata[j++];

    /* disregard space and newline */
    if (c == ' ' || c == '\n')
      continue;

    /* disregard non-hex alpha characters */
    if (c > 'f' || (c > 'F' && c < 'a'))
      continue;

    /* disregard non-numeric characters */
    if (c < '0' || (c > '9' && c < 'A'))
      continue;

    if (c > 'a'-1) /* lower case letters */
    {
      c -= 'a';
      c += 10;
    }
    else
    {
      if (c > 'A'-1) /* upper case letters */
      {
        c -= 'A';
        c += 10;
      }
      else
        c -= '0'; /* numbers */
    }

    /* c is now the numeric representation of the character */

    if (fst) /* first in a pair of characters */
    {
      ch = c * 16;
      fst = 0;
    }
    else
    {
      /* second in a pair */
      ch += c;
      outdata[k++] = ch;
      fst = 1;
    }
  }
  return k;
}

get_message(FILE *fp, char *buffer)
{
  int messlen;

  if (fp == stdin)
  {
    printf("Enter message data:\n");fflush(stdout);
    if (fgets(buffer, MSGSIZ-1, fp) != NULL)
    {
      messlen = strlen(buffer);       /* length without null        */
      if (buffer[messlen-1] == '\n')  /* last char is a new-line    */
      {
        buffer[messlen-1]  = '\0';    /* replace new-line with null */
        --messlen;                    /* reduce buffer length       */
      }
    }
    else 
      messlen = -1;        /* treat EOF same as null line         */
  }
  else
  {
    messlen = fread(buffer,1,MSGSIZ,fp);
  }

  return(messlen);
}

#define PRT_OFFSET  37
#define PRT_LEN     16

dbgprt(FILE *fp, unsigned char *ch, int size)
{
  int j,k,c;
  char prtbuf[101];
  char chp[3];

  c = 0;
  k = 0;
  memset(prtbuf,' ',100);

  for (j = 0; j < size; j++)
  {
    if (c >= 16)
    {
      *(prtbuf + PRT_LEN + PRT_OFFSET) = '\0';
      fprintf(fp,"   %s\n",prtbuf);
      c = 0;
      k = 0;
      memset(prtbuf,' ',100);
    }
    if (c && c%4 == 0)
    {
      *(prtbuf+k) = ' ';
      k++;
    }

    sprintf(chp,"%02X",*ch);
    strncpy(prtbuf+k,chp,2);
    k += 2;

    if ((*ch >= '0' && *ch <= 'z') || *ch == ' ')
    {
      *(prtbuf+PRT_OFFSET+c) = *ch;
    }
    else
    {
      *(prtbuf+PRT_OFFSET+c) = '.';
    }
    ch++;
    c++;
  }
  *(prtbuf + PRT_LEN + PRT_OFFSET) = '\0';
  fprintf(fp,"   %s\n",prtbuf);
}

usage(char *prog, int err)
{
  printf("Usage: %s\n",prog);
  printf("  Connection\n");
  printf("  ----------\n");
  printf("  -M = Queue Manager (mandatory)\n");
  printf("  -Q = Reply-to queue\n");
  printf("  -F = Reply output file (outf)\n");
  printf("  Actions\n");
  printf("  -------\n");
  printf("  -r = Register   Subscriber\n");
  printf("  -R = Register   Publisher\n");
  printf("  -d = Deregister Subscriber\n");
  printf("  -D = Deregister Publisher\n");
  printf("  -p = Publish\n");
  printf("  -P = Delete Publication\n");
  printf("  -U = Request Update\n");
  printf("  Data\n");
  printf("  ----\n");
  printf("  -s = stream, default SYSTEM.BROKER.DEFAULT.STREAM\n");
  printf("  -t = topic, (mandatory except Deregister with Deregister All)\n");
  printf("  -q = queue, (mandatory, except Delete Publication)\n");
  printf("  -f = filename for message data, Publish only, default stdin\n");
  printf("  Options\n");
  printf("  -------\n");
  printf("  -c = CorrelAsId text, max 24, space filled\n");
  printf("  -C = CorrelAsId hex chars, max 48, NULL filled\n");
  printf("  -a = Deregister All (Deregister Publisher/Subscriber Only)\n");
  printf("  -n = New Pubs Only (Register Subscriber Only)\n");
  printf("  -o = Pub on Req Only (Register Subscriber Only)\n");
  printf("  -i = Inform if Ret (Register Subscriber Only)\n");
  printf("  -k = Retain Publication (Publish Only)\n");

  switch (err)
  {
  case ERR_NOQMGR:
    printf("Queue Manager not specified\n");
    break;
  case ERR_NOQUEUE:
    printf("Queue not specified\n");
    break;
  case ERR_RETPUB:
    printf("Retain Pub option, action not Publish\n");
    break;
  case ERR_REGSUB:
    printf("New Pubs, Pub On Req, Inform if Ret option, action not Register Subscriber\n");
    break;
  case ERR_DEREG:
    printf("Deregister All option, action not Deregister\n");
    break;
  case ERR_TOPIC:
    printf("Topic not specified\n");
    break;
  }
}
